if (Symbol && !Symbol.asyncIterator) {
    Symbol.asyncIterator = Symbol.for("__@@asyncIterator__");
}
class ItemIterator {
    constructor(paginator) {
        this.paginator = paginator;
        this._iteratedCount = 0;
        this.lastResolved = Promise.resolve();
        this.pending = [];
    }
    /**
     * @inheritDoc
     */
    [Symbol.asyncIterator]() {
        return this;
    }
    /**
     * The capacity units consumed by the Scan operation. The data returned
     * includes the total provisioned throughput consumed, along with statistics
     * for the table and any indexes involved in the operation. ConsumedCapacity
     * is only returned if the ReturnConsumedCapacity parameter was specified.
     */
    get consumedCapacity() {
        return this.paginator.consumedCapacity;
    }
    /**
     * The number of items that have been iterated over.
     */
    get count() {
        return this._iteratedCount;
    }
    /**
     * @inheritDoc
     */
    next() {
        this.lastResolved = this.lastResolved.then(() => this.getNext());
        return this.lastResolved;
    }
    /**
     * Detaches the underlying paginator from this iterator and returns it. The
     * paginator will yield arrays of unmarshalled items, with each yielded
     * array corresponding to a single call to the underlying API. As with the
     * underlying API, pages may contain a variable number of items or no items,
     * in which case an empty array will be yielded.
     *
     * Calling this method will disable further iteration.
     */
    pages() {
        // Prevent the iterator from being used further and squelch any uncaught
        // promise rejection warnings
        this.lastResolved = Promise.reject(new Error('The underlying paginator has been detached from this iterator.'));
        this.lastResolved.catch(() => { });
        return this.paginator;
    }
    /**
     * @inheritDoc
     */
    return() {
        // Prevent any further use of this iterator
        this.lastResolved = Promise.reject(new Error('Iteration has been manually interrupted and may not be resumed'));
        this.lastResolved.catch(() => { });
        // Clear the pending queue to free up memory
        this.pending.length = 0;
        return this.paginator.return().then(doneSigil);
    }
    /**
     * The number of items evaluated, before any ScanFilter is applied. A high
     * scannedCount value with few, or no, Count results indicates an
     * inefficient Scan operation. For more information, see Count and
     * ScannedCount in the Amazon DynamoDB Developer Guide.
     */
    get scannedCount() {
        return this.paginator.scannedCount;
    }
    getNext() {
        if (this.pending.length > 0) {
            this._iteratedCount++;
            return Promise.resolve({
                value: this.pending.shift(),
                done: false
            });
        }
        return this.paginator.next().then(({ done, value }) => {
            if (done) {
                return { done };
            }
            this.pending.push(...value.Items || []);
            return this.getNext();
        });
    }
}
function doneSigil() {
    return { done: true };
}

/**
 * @internal
 */
function mergeConsumedCapacities(a, b) {
    if (a || b) {
        a = a || {};
        b = b || {};
        if ((a.TableName && b.TableName) && a.TableName !== b.TableName) {
            throw new Error('Consumed capacity reports may only be merged if they describe the same table');
        }
        return {
            TableName: a.TableName || b.TableName,
            CapacityUnits: (a.CapacityUnits || 0) + (b.CapacityUnits || 0),
            Table: mergeCapacities(a.Table, b.Table),
            LocalSecondaryIndexes: mergeCapacityMaps(a.LocalSecondaryIndexes, b.LocalSecondaryIndexes),
            GlobalSecondaryIndexes: mergeCapacityMaps(a.GlobalSecondaryIndexes, b.GlobalSecondaryIndexes),
        };
    }
}
function mergeCapacities(a, b) {
    if (a || b) {
        return {
            CapacityUnits: ((a && a.CapacityUnits) || 0) +
                ((b && b.CapacityUnits) || 0),
        };
    }
}
function mergeCapacityMaps(a, b) {
    if (a || b) {
        const out = {};
        a = a || {};
        b = b || {};
        const keys = new Set();
        for (const map of [a, b]) {
            for (const indexName of Object.keys(map)) {
                keys.add(indexName);
            }
        }
        for (const key of keys) {
            out[key] = mergeCapacities(a[key], b[key]);
        }
        return out;
    }
}

if (Symbol && !Symbol.asyncIterator) {
    Symbol.asyncIterator = Symbol.for("__@@asyncIterator__");
}
class DynamoDbPaginator {
    constructor(limit) {
        this.limit = limit;
        this._count = 0;
        this._scannedCount = 0;
        this.lastResolved = Promise.resolve();
    }
    /**
     * @inheritDoc
     */
    [Symbol.asyncIterator]() {
        return this;
    }
    /**
     * @inheritDoc
     */
    get consumedCapacity() {
        return this._consumedCapacity;
    }
    /**
     * @inheritDoc
     */
    get count() {
        return this._count;
    }
    /**
     * Get the LastEvaluatedKey of the last result page yielded by this
     * paginator or undefined if the scan has already been exhausted.
     */
    get lastEvaluatedKey() {
        return this._lastKey;
    }
    /**
     * @inheritDoc
     */
    next() {
        this.lastResolved = this.lastResolved.then(() => {
            if (this.count >= (this.limit === undefined ? Infinity : this.limit)) {
                return { done: true };
            }
            return this.getNext().then(({ done, value }) => {
                if (value && !done) {
                    this._lastKey = value.LastEvaluatedKey;
                    this._count += (value.Items || []).length;
                    this._scannedCount += (value.ScannedCount || 0);
                    this._consumedCapacity = mergeConsumedCapacities(this._consumedCapacity, value.ConsumedCapacity);
                }
                return { value, done };
            });
        });
        return this.lastResolved;
    }
    /**
     * @inheritDoc
     */
    return() {
        // Prevent any further use of this iterator
        this.lastResolved = Promise.reject(new Error('Iteration has been manually interrupted and may not be resumed'));
        this.lastResolved.catch(() => { });
        return Promise.resolve({ done: true });
    }
    /**
     * @inheritDoc
     */
    get scannedCount() {
        return this._scannedCount;
    }
    getNextPageSize(requestedPageSize) {
        if (this.limit === undefined) {
            return requestedPageSize;
        }
        return Math.min(requestedPageSize === undefined ? Infinity : requestedPageSize, this.limit - this.count);
    }
}

class ScanPaginator extends DynamoDbPaginator {
    constructor(client, input, limit) {
        super(limit);
        this.client = client;
        this.nextRequest = Object.assign({}, input, { Limit: this.getNextPageSize(input.Limit) });
    }
    getNext() {
        if (this.nextRequest) {
            return this.client.scan(Object.assign({}, this.nextRequest, { Limit: this.getNextPageSize(this.nextRequest.Limit) }))
                .promise()
                .then(output => {
                if (this.nextRequest && output.LastEvaluatedKey) {
                    this.nextRequest = Object.assign({}, this.nextRequest, { ExclusiveStartKey: output.LastEvaluatedKey });
                }
                else {
                    this.nextRequest = undefined;
                }
                return Promise.resolve({
                    value: output,
                    done: false
                });
            });
        }
        return Promise.resolve({ done: true });
    }
}

if (Symbol && !Symbol.asyncIterator) {
    Symbol.asyncIterator = Symbol.for("__@@asyncIterator__");
}
class ParallelScanPaginator {
    constructor(client, input, scanState = nullScanState(input.TotalSegments)) {
        this.pending = [];
        this.lastResolved = Promise.resolve();
        const { TotalSegments } = input;
        if (scanState.length !== TotalSegments) {
            throw new Error(`Parallel scan state must have a length equal to the number of `
                + `scan segments. Expected an array of ${TotalSegments} but`
                + `received an array with ${scanState.length} elements.`);
        }
        this.iterators = new Array(TotalSegments);
        for (let i = 0; i < TotalSegments; i++) {
            const iterator = new ScanPaginator(client, Object.assign({}, input, { Segment: i, ExclusiveStartKey: scanState[i].LastEvaluatedKey }));
            this.iterators[i] = iterator;
            // If the segment has not been initialized or a pagination token has
            // been received, request the next page.
            if (!scanState[i].initialized || scanState[i].LastEvaluatedKey) {
                this.refillPending(iterator, i);
            }
        }
        this._scanState = [...scanState];
    }
    /**
     * @inheritDoc
     */
    [Symbol.asyncIterator]() {
        return this;
    }
    /**
     * @inheritDoc
     */
    get consumedCapacity() {
        return this.iterators.reduce((merged, paginator) => mergeConsumedCapacities(merged, paginator.consumedCapacity), undefined);
    }
    /**
     * @inheritDoc
     */
    get count() {
        return this.iterators.reduce((sum, paginator) => sum + paginator.count, 0);
    }
    /**
     * @inheritDoc
     */
    next() {
        this.lastResolved = this.lastResolved.then(() => this.getNext());
        return this.lastResolved;
    }
    async getNext() {
        if (this.pending.length === 0) {
            return doneSigil$1();
        }
        // Grab the next available result from any segment.
        const { iterator, result: { value, done }, segment, } = await Promise.race(this.pending.map(pending => pending.result));
        // Update the scan state for this segment. This will either be the last
        // evaluated key (for an unfinished segment) or undefined (for a
        // completed segment).
        this._scanState[segment] = {
            initialized: true,
            LastEvaluatedKey: value && value.LastEvaluatedKey,
        };
        // Remove the result from the pending set.
        for (let i = this.pending.length - 1; i >= 0; i--) {
            if (this.pending[i].iterator === iterator) {
                this.pending.splice(i, 1);
            }
        }
        // If the iterator is not finished, add its next result to the pending
        // set.
        if (!done) {
            this.refillPending(iterator, segment);
            return { value, done };
        }
        else {
            // If a segment has finished but there are still outstanding
            // requests, recur. A done sigil will be returned when the pending
            // queue is empty.
            return this.getNext();
        }
    }
    /**
     * @inheritDoc
     */
    async return() {
        this.pending.length = 0;
        return Promise.all(this.iterators.map(iterator => iterator.return()))
            .then(doneSigil$1);
    }
    /**
     * @inheritDoc
     */
    get scannedCount() {
        return this.iterators.reduce((sum, paginator) => sum + paginator.scannedCount, 0);
    }
    /**
     * A snapshot of the current state of a parallel scan. May be used to resume
     * a parallel scan with a separate paginator.
     */
    get scanState() {
        return [...this._scanState];
    }
    refillPending(iterator, segment) {
        // Use .push to reorder segments within the array of pending results.
        // Promise.race will iterate over the array of pending results until a
        // resolved promise is found and therefore will naturally favor promises
        // towards the head of the queue. Removing resolved segments and sending
        // them to the back of the line will keep this implementation detail
        // from creating hot and cold scan segments.
        this.pending.push({
            iterator: iterator,
            result: iterator.next()
                .then(result => ({ iterator, result, segment })),
        });
    }
}
function doneSigil$1() {
    return { done: true };
}
/**
 * `Array.prototype.fill` is not available in IE, so a loop is used instead
 */
function nullScanState(length) {
    const target = new Array(length);
    for (let i = 0; i < length; i++) {
        target[i] = { initialized: false };
    }
    return target;
}

class ParallelScanIterator extends ItemIterator {
    constructor(client, input, scanState) {
        super(new ParallelScanPaginator(client, input, scanState));
    }
}

class QueryPaginator extends DynamoDbPaginator {
    constructor(client, input, limit) {
        super(limit);
        this.client = client;
        this.nextRequest = Object.assign({}, input);
    }
    getNext() {
        if (this.nextRequest) {
            return this.client.query(Object.assign({}, this.nextRequest, { Limit: this.getNextPageSize(this.nextRequest.Limit) }))
                .promise()
                .then(output => {
                if (this.nextRequest && output.LastEvaluatedKey) {
                    this.nextRequest = Object.assign({}, this.nextRequest, { ExclusiveStartKey: output.LastEvaluatedKey });
                }
                else {
                    this.nextRequest = undefined;
                }
                return Promise.resolve({
                    value: output,
                    done: false
                });
            });
        }
        return Promise.resolve({ done: true });
    }
}

class QueryIterator extends ItemIterator {
    constructor(client, input, limit) {
        super(new QueryPaginator(client, input, limit));
    }
}

class ScanIterator extends ItemIterator {
    constructor(client, input, limit) {
        super(new ScanPaginator(client, input, limit));
    }
}

export { ParallelScanIterator, ParallelScanPaginator, QueryIterator, QueryPaginator, ScanIterator, ScanPaginator };
